import time
from tqdm import tqdm
from itertools import count
from collections import defaultdict

def render_step(verbose, causes, non_causes, Cs):
    non_causes = [v for v in non_causes if not any([c <= v[3] for c in Cs])]
    if len(non_causes):
        if verbose == 2:
            non_causes = sorted(non_causes, key=sort_key)
            print("Number of causes found:", len(causes))
            print("Number of non-causes remaining:", len(non_causes))
            print("Best non-cause:")
            show_rule(non_causes[0])
            print("Worst non-cause:")
            show_rule(non_causes[-1])
        if verbose >= 3:
            print("Causes found this step:")
            show_rules(causes)
            print("Rules passed for next step:")
            show_rules(non_causes)
    else:
        if verbose >= 2:
            print("No rule available")

def get_sets(rule, actual_values):
    C = set()
    W = set()
    for feature, value in rule:
        if actual_values[feature] != value:
            C.add(feature)
        else:
            W.add(feature)
    return C, W

def sort_key(rule_values):
    _, rule_output, rule_score, C, W = rule_values
    return (rule_score, len(C), C, len(W), W)

def format_value(value):
    if isinstance(value, float):
        return f"{float(value):.2f}"
    elif isinstance(value, int):
        return f"{int(value)}"
    else:
        return f"{value}"

def get_rule_desc(rule_values, show_score=False):
    rule, output, score, C, W = rule_values
    dim2value = dict(rule)
    C = {c:format_value(dim2value[c]) for c in C}
    W = {w:format_value(dim2value[w]) for w in W}
    if show_score: 
        if isinstance(score, tuple):
            return f"{C=}, {W=}, {output=}, {score=}"
        return f"{C=}, {W=}, {output=}, {score=:.3f}"
    return f"{C=}, {W=}"

def show_rule(rule_values):
    print(get_rule_desc(rule_values, True))
    
def show_rules(rule_values):
    for r_values in rule_values:
        show_rule(r_values)

def is_minimal(e, E):
    return not any([other[3] < e[3] for other in E])

def minimal_merge(E1, E2):
    E = [e for e in E1 if is_minimal(e, E2)] + [e for e in E2 if is_minimal(e, E1)]
    return remove_duplicates(E)

def filter_minimality(E):
    # Remove causes with strict subsets that are causes
    E = [e for e in E if is_minimal(e, E)]

    # If there are equalities, keep the ones with smallest W and best score
    E = remove_duplicates(E)
    return E

def remove_duplicates(E):
    Cs = defaultdict(lambda: [])
    for e in E:
        Cs[tuple(e[3])].append(e)
    causes = []
    for cands in Cs.values():
        best = min(cands, key=lambda e: (len(e[4]),e[2]))
        causes.append(best)
    return causes
    

def get_initial_rules(V, D, v):
    rules = []
    for actual_value, feature, domain in zip(v, V, D):
        for value in domain:
            if actual_value != value:
                rules.append(((feature, value),))
    return rules

def filter_instance(V, D, v, I):
    return zip(*[(variable, domain, value) for variable, domain, value in zip(V, D, v) if variable in I])

def get_rules(previous_rules, V, D, v, actual_values, Cs=[], R=[], verbose=False):
    # Build new rules on top of the previous ones
    # The previous rules are not valid (i.d. they do not define causes)
    # if verbose: print(previous_rules, V, D, v, Cs, actual_values)
    if previous_rules is None: 
        rules = get_initial_rules(V, D, v)
        return [beam + R for beam in sorted(rules)]
    new_rules = set()

    # Iterate through the previous rules
    for rule in tqdm(previous_rules, disable=not verbose): # Complexity: O(1)
        C, W = get_sets(rule, actual_values)
        for actual_value, feature, domain in zip(v, V, D): # Complexity: O(n)
            # Don't consider features already in rule
            if feature in C|W:
                continue
                
            # Don't consider the rule if it is not minimal
            non_minimal_c = any([c <= C|{feature} for c in Cs]) # Complexity O(n)
            

            # Add new rules with the feature
            for value in domain:
                # Check for minimality if we add a new variable to C
                if value != actual_value and non_minimal_c:
                    continue
                # Build the rule
                new_rule = rule + ((feature, value),)
                # Add the new rule to the next rules
                new_rules.add(tuple(sorted(new_rule)))
    return [beam + R for beam in sorted(new_rules)]

def get_next_beam(non_causes, beam_size, Cs):
    non_causes = [v for v in non_causes if not any([c <= v[3] for c in Cs])]
    # Score and sort the remaining
    non_causes = sorted(non_causes, key=sort_key)
    # Filter the top-b
    if beam_size != -1:
        non_causes = non_causes[:beam_size]
    # Keep only the interventions to build next ones
    beam = [rule_value[0] for rule_value in non_causes]
    return beam

def split_rules(beam, cf_values, actual_values, epsilon):
    causes, non_causes = [], []
    for rule, (cf_output, cf_score) in zip(beam, cf_values):
        C, W = get_sets(rule, actual_values)
        rule_value = (rule, float(cf_output), float(cf_score), C, W)

        # Save causes and keep n best non-causes for next step
        if cf_output < epsilon: 
            causes.append(rule_value)
        else:
            non_causes.append(rule_value)
    return causes, non_causes

def check_early_stop(beam, early_stop, all_causes, max_time, init_time):
    if not len(beam): 
        return True
    if early_stop and len(all_causes): 
        return True
    if max_time is not None and time.time()-init_time > max_time: 
        return True
    return False

def do_simulation(simulation, cache, beam):
    if cache is None: 
        return simulation(beam)
    cached_results = []
    non_cached_beam = []
    for e in beam:
        if e in cache:
            cached_results.append(cache[e])
        else:
            non_cached_beam.append(e)
    cf_values = simulation(non_cached_beam)
    cache |= dict(zip(non_cached_beam, cf_values))
    return cached_results + cf_values

# def test_empty(simulation):
#     beam = [()]
#     cf_values = simulation(beam)

def beam_search(
    v, D, simulation, V, # SCM
    max_steps=5, beam_size=10, epsilon=.05, early_stop=False, max_time=None, # Parameters
    R=None, Cs=None, cache=None, minimality=True, I=None, # Additional parameters when running for sub-instance
    verbose=0, 
    ):
    # verbose: 
    #  = 1 -> best cause at the end, tqdm for steps
    #  >= 2 -> removes step tqdm, adds step header + number of cause found + best and worse non causes
    #  >= 3 -> adds all causes + tqdm for get_rules
    
    all_causes = []
    if R is None: 
        R = tuple()
    actual_values = dict(zip(V, v))
    if I is not None: 
        V, D, v = filter_instance(V,D, v, I)
    # print(I, V)
    if not minimality: 
        full_interventions = []
    init_time = time.time()
    beam = None
    if Cs is None: 
        Cs = []
    if max_steps == -1 or max_steps is None: 
        iterator = count(start=1, step=1)
    else: 
        iterator = range(1,max_steps+1)

    # if test_empty(simulation): iterator = range(0,-1)

    for t in tqdm(iterator, disable=(verbose!=1)):
        # Render the step
        if verbose >= 2: 
            print(f"{f'Step {t}':=^30}")
            
        # Create the rules for step t base on the ones from t-1, we use the initial ones if t==1
        beam = get_rules(beam, V, D, v, actual_values, 
                          Cs=Cs if minimality else [], 
                          R=R, verbose=verbose >= 3)

        # Check for early stop
        if check_early_stop(beam, early_stop, all_causes, max_time, init_time):
            break
            
        # Render how many nodes will be evaluated
        if verbose >= 2: 
            print(f"Evaluating {len(beam)} rules")

        # Evaluate the rules using the simulation 
        cf_values = do_simulation(simulation, cache, beam)
        # Build the tuples of rule values
        AC2, non_AC2 = split_rules(beam, cf_values, actual_values, epsilon)

        # Filter causes to keep only minimal ones
        causes = filter_minimality(AC2)

        # Save minimal causes
        if minimality:
            all_causes += causes
        # Search for all counterfactual interventions:
        else:
            # Filter non-minimal propagated causes
            all_causes = minimal_merge(all_causes, causes)
            # Save all counterfactual interventions
            full_interventions += AC2
            # Remove duplicates but not non-minimal interventions
            # full_interventions = remove_duplicates(full_interventions)

        # Save the cause sets
        for rule_value in causes:
            Cs.append(rule_value[3])

        # Build next beam
        if minimality:
            beam = get_next_beam(non_AC2, beam_size, Cs)
        else:
            beam = get_next_beam(non_AC2 + AC2, beam_size, [])
        # Render step output
        render_step(verbose, causes, non_AC2, Cs)

    # Render final result
    if verbose:
        print(f"----> Found {len(all_causes)} causes.")
    if not minimality: 
        return all_causes, full_interventions
    return all_causes



from .lucb import lucb
import numpy as np
import random
from collections import defaultdict

class SystemModel:
    def __init__(self, phi=None, psi=None):
        self.n_calls = 0
        self.psi = psi
        self.phi = phi

    def __call__(self, u:list, e: list[tuple]) -> list:
        # Compute a state of the system model given context u and intervention e
        raise NotImplementedError

    def evaluate_batch(self, u, E, N=1):
        # Compute the values of phi and psi for context u and N times for each intervention e in E
        ret = []
        for e in E:
            for _ in range(N):
                ret.append((
                    self.phi(self(u,e)), self.psi(self(u,e))
                ))
        return ret


class SuzzyExampleSystemModel(SystemModel):
    # A model for the rock throwing example, with variables ST, BT, SH, BH, BS
    def __init__(self):
        super().__init__(
            phi=lambda s: s[-1], 
            psi=lambda s: sum(s)
        )

    def __call__(self, u:list, e: list[tuple]) -> list:
        st, bt = u
        e = dict(e)
        ST = e.get("ST", st)
        BT = e.get("BT", bt)
        SH = e.get("SH", ST)
        BH = e.get("BH", int(BT and not SH))
        BS = int(SH or BH)
        self.n_calls += 1
        return [ST,BT,SH,BH,BS]

class BaseNumpyModel(SystemModel):
    sub_N = 100_000 # used to chunk large batches
    def __init__(self, V, phi=None, psi=None, dtype=None):
        SystemModel.__init__(self, phi=phi, psi=psi)
        if dtype is None: self.dtype = bool
        else: self.dtype = dtype
        if self.phi is None: 
            self.phi = lambda s: s[:,-1].astype(int)
        if self.psi is None: 
            self.psi = lambda s: np.sum(s, axis=1) - 1
        self.dim2id = dict(zip(V, range(len(V))))
        self.reset_state()

    def reset_state(self):
        self.Es = None
        self.batch = None
        self.S = None

    def __getitem__(self, var: set):
        return self.S[:,self.dim2id[var]]
            
    def __setitem__(self, var: str, F_value):
        if self.batch: 
            self.S[:,self.dim2id[var]] = F_value
            if var in self.Es:
                for h_slice, value in self.Es[var]:
                    self.S[h_slice,self.dim2id[var]] = value
        else: 
            self.S[:,self.dim2id[var]] = self.Es.get(var, F_value)
            
    def __call__(self, u, e):
        self.batch = False
        self.S = np.zeros((1, len(self.dim2id)), dtype=self.dtype)
        self.Es = dict(e)
        self.simulate(u)
        S = self.S
        self.n_calls += S.shape[0]
        self.reset_state()
        return S.flatten().tolist()

    def evaluate_batch(self, u, E, N=1):
        self.batch = True
        out = []
        for i in range(0, len(E), self.sub_N):
            sub_E = E[i*self.sub_N:(i+1)*self.sub_N]
            self.S = np.zeros((len(sub_E)*N, len(self.dim2id)), dtype=self.dtype)
            self.Es = defaultdict(lambda: [])
            for i, e in enumerate(sub_E):
                for var, value in e:
                    self.Es[var].append((slice(i*N,(i+1)*N),value))
            self.simulate(u)
            out.append(np.array([self.phi(self.S), self.psi(self.S)]).T)
            self.n_calls += self.S.shape[0]
        if not len(out):
            return 
        return np.vstack(out)

class NoisyNumpyModel(BaseNumpyModel):
    def __init__(self, V, t, phi=None, psi=None, dtype=None, rng=None):
        BaseNumpyModel.__init__(self, V, phi=phi, psi=psi, dtype=dtype)
        self.t = t
        self.rng = rng

    def __setitem__(self, var, F_value):
        if self.batch: 
            self.S[:,self.dim2id[var]] = F_value
            if self.t > 0:
                if self.rng is None:
                    ids = np.random.rand(self.S.shape[0]) < self.t
                else:
                    ids = self.rng.random (self.S.shape[0]) < self.t
                self.S[ids,self.dim2id[var]] = 1 - self.S[ids,self.dim2id[var]]
            if var in self.Es:
                for h_slice, value in self.Es[var]:
                    self.S[h_slice,self.dim2id[var]] = value
        else: 
            if self.rng is None:
                rd = np.random.rand()
            else:
                rd = self.rng.random()
            if self.t > 0 and rd < self.t: 
                F_value = 1 - F_value
            self.S[:,self.dim2id[var]] = self.Es.get(var, F_value)

class AverageNumpyModel(NoisyNumpyModel):
    def __init__(self, V, t, N, seed=None, phi=None, psi=None, dtype=None, rng=None):
        NoisyNumpyModel.__init__(self, V, t, phi=phi, psi=psi, dtype=dtype, rng=rng)
        self.seed=seed
        self.N = N

    def evaluate_batch(self, u, E, N=1):
        if self.seed is not None:
            np.random.seed(self.seed)
            random.seed(self.seed)
        out = super().evaluate_batch(u, E, self.N)
        return out.reshape(len(E), self.N, 2).mean(axis=1)

class LUCBNumpyModel(NoisyNumpyModel):
    def __init__(self, V, t, lucb_params, phi=None, psi=None, dtype=None, rng=None):
        NoisyNumpyModel.__init__(self, V, t, phi=phi, psi=psi, dtype=dtype, rng=rng)
        self.lucb_params = lucb_params

    def evaluate_batch(self, u, E, N=1):
        def evaluator(E, N):
            return NoisyNumpyModel.evaluate_batch(self, u, E, N)
        return lucb(evaluator, E, **self.lucb_params)


from collections import deque
from itertools import combinations, islice
from collections.abc import Iterable, Iterator

from .mbs import beam_search, remove_duplicates, minimal_merge


def merge_set_lists(list1:list[set], list2:list[set])->list[set]:
    # Convert each set in the lists to a frozenset and add to a set to remove duplicates
    unique_frozensets = set()
    for s in list1:
        unique_frozensets.add(frozenset(s))
    for s in list2:
        unique_frozensets.add(frozenset(s))
    # Convert the frozensets back to sets
    merged_list = [set(fs) for fs in unique_frozensets]
    return merged_list

def make_beam_search(v:list[object], D:list[list], V:list[str], K:'constraint', minimality:bool, 
                     **kargs) -> tuple['minimal MBS output', 'non-minimal MBS output']:
    # Use K to update D (using W_0) and build new V=W_0 U I
    I, W_0, R_C, r_C, R_W = K
    I = I|W_0
    R = tuple(r_C.items())
    R += tuple([(var, value) for var, value in zip(V,v) if var in R_W])
    D = [D[i] if var not in W_0 else [value] for i, (var, value) in enumerate(zip(V,v))]
    
    if minimality:
        E = beam_search(v=v, D=D, V=V, I=I, R=R, **kargs)
        full_E = None
    else:
        E, full_E = beam_search(v=v, D=D, V=V, I=I, R=R, minimality=False, **kargs)
        full_E = remove_duplicates(full_E)
    return E, full_E

def subsets(s:Iterable, n:int=None) -> Iterator[tuple]:
    s = list(s)
    if n is None:
        n = len(s)
    n = min(n + 1, len(s) + 1)
    for size in range(1, n):
        for subset in combinations(s, size):
            yield set(subset)

def CH(var:str, PA:dict[str:set[str]])->set[str]:
    return {child for child, parents in PA.items() if var in parents}

def CH_set(S:set[str], PA:dict[str:set[str]])->set[str]:
    chs = [CH(var, PA) for var in S]
    return set.union(*chs) if chs else set()

def PA_set(S:set[str], PA:dict[str:set[str]])->set[str]:
    pas = [set(PA[var]) for var in S]
    return set.union(*pas) if pas else set()

def desc(S:set[str], PA:dict[str:set[str]])->set[str]:
    ch = CH_set(S, PA)
    des = set()
    while ch:
        des |= ch
        ch = CH_set(ch, PA)
    return des

def anc(S:set[str], PA:dict[str:set[str]])->set[str]:
    pa = PA_set(S, PA)
    an = set()
    while pa:
        an |= pa
        pa = PA_set(pa, PA)
    return an

def expand(C:set, e:dict, W:set, S:set, PA:dict[str:set[str]]) -> tuple[set, set, set, dict,set]:
    I = PA_set(S, PA) - C
    W_0 = (desc(I, PA) - I) & (anc(C, PA) - C)
    R_C = C - S
    r_C = {var:e[var] for var in R_C}
    R_W = W | (desc(I, PA) - I - (desc(C, PA) | C | anc(C, PA)))
    return I, W_0, R_C, r_C, R_W

def check_memory(memory, K):
    I, W_0, R_C, r_C, R_W = K
    for I_ref, W_0_ref, R_C_ref, r_C_ref, R_W_ref in memory:
        W_0 -= W_0_ref
        R_C = {var for var in R_C if var not in R_C_ref or r_C[var] != r_C_ref[var]}
        R_W -= R_W_ref
        if I_ref >= I | W_0 | R_C | R_W:
            # print(I_ref, W_0_ref, R_C_ref, r_C_ref, R_W_ref, ">=", K)
            return False
    return True

def iterative_identification(v, D, simulation, V, dag, PA_T, cache_size=-1, **kargs):
    PA = dag
    minimality = max(map(len,D)) <= 2
    verbose = kargs.get("verbose", 0)
    early_stop = kargs.get("early_stop", False)
    kargs["simulation"] = simulation
    K_0 = (set(PA_T), set(), set(), dict(), set())
    queue = deque([K_0])
    cache = dict() if cache_size >= 0 else None
    memory = []
    ret = []
    Cs = []
    while queue:
        # Set up node
        if verbose: 
            print(f"{len(queue)=}")
        K = queue.popleft()
        while cache and len(cache) > cache_size:
            item = next(iter(cache))
            del cache[item]
        if cache_size >=0: 
            cache = dict(islice(cache.items(), cache_size))
        if verbose: 
            print(f"{K=}")
        
        # Evaluate node
        E, full_E = make_beam_search(v, D, V, K, minimality, cache=cache, Cs=Cs, **kargs)
        
        Cs = merge_set_lists([e[3] for e in E], Cs)
        if early_stop and E: 
            return E
        ret = minimal_merge(E, ret)

        # Expand node
        if not minimality: 
            E = full_E
        for e in E:
            C, W = e[3], e[4]
            e = dict(e[0])
            for S in subsets(C):
                K = expand(C,dict(e),W,S, PA)
                if not len(K[0]) or not check_memory(memory, K): 
                    continue # no free variable for interventions or K already tested
                if verbose: 
                    print(f"  {C=} -> {S=} -> {K=}")
                queue.append(K)
                memory.append(K)
        
        if verbose: 
            print("==========")
    return ret



import numpy as np
from tqdm import tqdm
import random

# This file is inspired by https://github.com/marcotcr/anchor

def kl_bernoulli(p, q):
    p = np.clip(p, 1e-10, 1 - 1e-10)
    q = np.clip(q, 1e-10, 1 - 1e-10)
    return (p * np.log(p / q) + (1 - p) * np.log((1 - p) / (1 - q)))

def dup_bernoulli(mean, var, n, beta, max_iter=20, tol=1e-4):
    # print(f"{mean=}, {var=}, {n=}, {beta=}, {max_iter=}, {tol=}")
    level = beta / n
    valid_ub = (level > 0) & (mean < 1 - tol)
    ub = np.ones_like(mean)

    if np.any(valid_ub):
        lm_ub = mean[valid_ub].copy()
        um_ub = np.minimum(1.0, mean[valid_ub] + np.sqrt(level[valid_ub] / 2.))
    
        for _ in range(max_iter):
            qm = (um_ub + lm_ub) / 2.0
            kl = kl_bernoulli(mean[valid_ub], qm)
            above = kl > level[valid_ub]
            um_ub = np.where(above, qm, um_ub)
            lm_ub = np.where(above, lm_ub, qm)
            
            if np.all((um_ub - lm_ub) < tol):
                break
        ub[valid_ub] = um_ub
    return ub

def dlow_bernoulli(mean, var, n, beta, max_iter=20, tol=1e-4):
    level = beta / n
    valid_lb = (level > 0) & (mean > tol)
    lb = np.zeros_like(mean)
    
    if np.any(valid_lb):
        um_lb = mean[valid_lb].copy()
        lm_lb = np.maximum(0.0, mean[valid_lb] - np.sqrt(level[valid_lb] / 2.))
        
        for _ in range(max_iter):
            qm = (um_lb + lm_lb) / 2.0
            kl = kl_bernoulli(mean[valid_lb], qm)
            above = kl > level[valid_lb]
            lm_lb = np.where(above, qm, lm_lb)
            um_lb = np.where(above, um_lb, qm)
            
            if np.all((um_lb - lm_lb) < tol):
                break
        lb[valid_lb] = lm_lb
    return lb

def hoeffding_upper_bound(mean, var, n, beta):
    ub = mean + np.sqrt(beta / (2 * n))
    ub = np.minimum(1.0, ub)
    return ub

def hoeffding_lower_bound(mean, var, n, beta):
    lb = mean - np.sqrt(beta / (2 * n))
    lb = np.maximum(0.0, lb)
    return lb

def bernstein_lower_bound(mean, var, n, beta):
    eps = np.sqrt(2 * var * beta / n) + 2 * beta / (3 * n)
    return np.maximum(0.0, mean - eps)

def bernstein_upper_bound(mean, var, n, beta):
    eps = np.sqrt(2 * var * beta / n) + 2 * beta / (3 * n)
    return np.minimum(1.0, mean + eps)

def compute_beta_exact(n_arms, t, delta=0.1):
    return np.log(2 * n_arms * t**2 / delta)

def compute_beta_practical(n_arms, t, delta=0.1):
    return np.log(2 * n_arms / delta)

def compute_beta_usable(n_arms, t, delta=0.1):
    return np.log(2 / delta)

compute_beta = compute_beta_usable

def update_bound(stats, fnt, bound_id, arm_ids, mean_id, variance_id, n_id, beta):
    var = stats[variance_id,arm_ids] if variance_id is not None else None
    stats[bound_id, arm_ids] = fnt(stats[mean_id,arm_ids], var, stats[n_id,arm_ids], beta)


def lucb(evaluator, rules, beam_size, a=.05, beam_eps=.1, cause_eps=.01, non_cause_eps=.01, 
         max_iter=200, verbose=0, batch_size=10, init_batch_size=20, lucb_info=None, delta=.1, seed=None):

    if seed is not None:
        np.random.seed(seed)
        random.seed(seed)
    init_batch_size = max(1, init_batch_size)
    n_arms = len(rules) # Doing armed bandits with the rules to evaluate

    n_stats = 9
    phi_m, phi_ub, phi_lb, psi_m, psi_v, psi_M2, psi_ub, psi_lb, n = range(n_stats)
    stats = np.zeros((n_stats,n_arms), dtype=float) 
    stats[phi_ub] = 1.0
    stats[psi_ub] = 1.0
    
    # Utils function
    def action_arms(arms, bs=batch_size):
        # Compute values
        E = [rules[arm] for arm in arms]
        if not len(E): 
            return
        values_batchs = evaluator(E, bs)
        values_batchs = values_batchs.reshape(len(E), bs, -1)
        for arm, values_batch in zip(arms, values_batchs):
            # Update n 
            n_old = stats[n,arm]
            stats[n,arm] += bs
            # Compute batch stats
            mean_batch = np.mean(values_batch, axis=0)
            M2_batch = np.sum((values_batch[:,1] - mean_batch[1])**2) # Compute M2 only for psi
            # Update phi
            stats[phi_m,arm] += (mean_batch[0] - stats[phi_m, arm]) * bs / stats[n,arm]
            # Update psi
            stats[psi_m,arm] += (mean_batch[1] - stats[psi_m, arm]) * bs / stats[n,arm]
            delta = mean_batch[1] - stats[psi_m, arm]
            stats[psi_m,arm] += delta * bs / stats[n,arm]
            stats[psi_M2,arm] += M2_batch + delta**2 * n_old * bs / stats[n,arm]
            stats[psi_v,arm] = stats[psi_M2,arm] / stats[n,arm]
            
    def find_unsure_arms():
        cancel_ids = stats[phi_m] < a
        # Cause candidates
        cause_ids = np.argwhere(cancel_ids).flatten()
        if not cause_ids.size:
            cause_overlap = np.array([], dtype=int)
        else:
            ids_overlapping = stats[phi_ub, cause_ids] >= a + cause_eps
            cause_overlap = cause_ids[ids_overlapping]
        # Non-cause candidates
        non_cause_ids = np.argwhere(~cancel_ids).flatten()
        if not non_cause_ids.size:
            non_cause_overlap = np.array([], dtype=int)
        else:
            ids_overlapping = stats[phi_lb, non_cause_ids] <= a - non_cause_eps
            non_cause_overlap = non_cause_ids[ids_overlapping]
        # Beam overlap
        sorted_non_cause_ids = sorted(non_cause_ids, key = lambda i: stats[psi_m,i])
        beam_ids = np.array(sorted_non_cause_ids[:beam_size])
        non_beam_ids = np.array(sorted_non_cause_ids[beam_size:])
        if not non_beam_ids.size or not beam_ids.size:
            return cause_overlap, non_cause_overlap, np.array([], dtype=int), np.array([], dtype=int)
        ut = beam_ids[np.argmax(stats[psi_ub,beam_ids])]
        lt = non_beam_ids[np.argmin(stats[psi_lb,non_beam_ids])]
        # Beam candidates: overlap if the upper bound in the beam is higher than the lower bound outside the beam
        ids_overlapping = stats[psi_ub, beam_ids] >= stats[psi_lb, lt] + beam_eps  
        beam_overlap = beam_ids[ids_overlapping]
        # Non-beam candidates: overlap if the lower bound outside the beam is lower than the upper bound in the beam
        ids_overlapping = stats[psi_lb, non_beam_ids] <= stats[psi_ub, ut] - beam_eps
        nonbeam_overlap = non_beam_ids[ids_overlapping]
        return cause_overlap, non_cause_overlap, beam_overlap, nonbeam_overlap
    
    # Initialization
    beam_bound = 1
    cause_bound = 1
    non_cause_bound = 1
    
    it = 1
    # Loop
    with tqdm(total=n_arms * max_iter, disable=not verbose) as pbar:
        while stats[n].sum() < n_arms * max_iter:
            beta = compute_beta(n_arms, it, delta)
            # Check necessary arms
            unsure_arms = find_unsure_arms()
            cause_overlap, non_cause_overlap, beam_overlap, nonbeam_overlap = unsure_arms
            
            if verbose:
                print(f"cause: {len(cause_overlap)}, non cause: {len(non_cause_overlap)}, beam: {len(beam_overlap)}, non beam: {len(nonbeam_overlap)}, ")
            
            # Pull arms
            action_arms(
                np.unique(np.hstack(unsure_arms)), 
                init_batch_size if it == 1 else batch_size
            )
            # Update bounds
            # Compute the upper and lower bound for psi of everyone after the first batch to stabilize computations
            if True:#it == 1: 
                update_bound(stats, bernstein_upper_bound, psi_ub, np.arange(n_arms),   psi_m, psi_v, n, beta)
                update_bound(stats, bernstein_lower_bound, psi_lb, np.arange(n_arms),   psi_m, psi_v, n, beta)
            else:
                update_bound(stats, bernstein_upper_bound, psi_ub, beam_overlap,      psi_m, psi_v, n, beta)
                update_bound(stats, bernstein_lower_bound, psi_lb, nonbeam_overlap,   psi_m, psi_v, n, beta)
            nonbeam_overlap = np.arange(n_arms)
            update_bound(stats, dup_bernoulli,         phi_ub, cause_overlap,     phi_m, None,  n, beta)
            update_bound(stats, dlow_bernoulli,        phi_lb, non_cause_overlap, phi_m, None,  n, beta)
            
            # Compute remaining overlap
            if not beam_overlap.size or not nonbeam_overlap.size: 
                beam_bound = 0
            else: 
                beam_bound = stats[psi_ub,beam_overlap].max() - stats[psi_lb,nonbeam_overlap].min()
                
            if not cause_overlap.size: 
                cause_bound = 0
            else: 
                cause_bound = stats[phi_ub,cause_overlap].max() - a
                
            if not non_cause_overlap.size: 
                non_cause_bound = 0
            else: 
                non_cause_bound = a - stats[phi_lb,non_cause_overlap].min()
            
            # Stop condition
            if beam_bound <= beam_eps and cause_bound <= cause_eps and non_cause_bound <= non_cause_eps: 
                if verbose > 1: 
                    print(f"Success: {beam_bound=:.4f} / {cause_bound=:.4f} / {non_cause_bound=:.4f})")
                break
            if cause_bound <= cause_eps and non_cause_bound <= non_cause_eps and (beam_size + (stats[phi_m] < a).sum()) >= n_arms:
                if verbose > 1:
                    print(f"All rules pass on to next state: {cause_bound=:.4f}, {non_cause_bound=:.4f}")
                break
            
            pbar.n = stats[n].sum()
            pbar.refresh()
            it += 1
        else:
            # Render how much we fail if we fail to reach the bound
            if verbose > 1: 
                print(f"Fail: {beam_bound=:.4f} / {cause_bound=:.4f} / {non_cause_bound=:.4f}")
            
    if verbose > 2:
        print(f"phi ub={stats[phi_ub].round(2)}")
        print(f"phi lb={stats[phi_lb].round(2)}")
        print(f"phi m={stats[phi_m].round(2)}")
        print(f"psi ub={stats[psi_ub].round(2)}")
        print(f"psi lb={stats[psi_lb].round(2)}")
        print(f"psi v={stats[psi_v].round(2)}")
        print(f"psi m={stats[psi_m].round(2)}")
        print(f"n_samples={stats[n]}")
    # print(f"Did {stats[n].sum()} calls (max={max_iter*len(rules)})")
    return stats[[phi_m,psi_m]].T
    # return [(stats[:,i], float(stats[phi_m,i]), float(stats[psi_m,i])) for i in range(n_arms)]



import sys
from pathlib import Path
import numpy as np
import random

from actualcauses import suzzy_example_scm, SCM

ROOT = Path(__file__).resolve().parents[1]
if str(ROOT) not in sys.path:
    sys.path.insert(0, str(ROOT))


from examples.custom_scm import ForestFireExampleSystemModel
from examples.vectorized_system_model import LampModel
from examples.stochastic_system_model import RockThrowingModel, AvgRockThrowingModel, LUCBRockThrowingModel
from examples.custom_heuristic import SuzzyCustomPsi
# import pytest


def test_example_1():
    dis_ff_example = SCM(
        V=("MD", "L", "FF"),
        U=("md", "l"),
        D=(0,1),
        u=(1,1),
        model=ForestFireExampleSystemModel(disjuctive=True),
        dag={"MD":[], "L":[], "FF":["MD", "L"]}
    )
    
    conj_ff_example = SCM(
        V=("MD", "L", "FF"),
        U=("md", "l"),
        D=(0,1),
        u=(1,1),
        model=ForestFireExampleSystemModel(disjuctive=False),
        dag={"MD":[], "L":[], "FF":["MD", "L"]}
    )
    dis_exp_causes = [{'MD', 'L'}]
    dis_ff_example.find_causes()
    dis_causes = dis_ff_example.causes
    assert set(map(frozenset, dis_exp_causes)) == set(map(frozenset, dis_causes))
    
    conj_exp_causes = [{'L'}, {'MD'}]
    conj_ff_example.find_causes()
    conj_causes = conj_ff_example.causes
    assert set(map(frozenset, conj_exp_causes)) == set(map(frozenset, conj_causes))

def test_example_2():
    scm_lamp = SCM(V=["A","B", "C", "L"], U=["a","b", "c"], D=[(-1,0,1)]*3+[(0,1)], 
                model=LampModel(V=["A","B", "C", "L"], dtype=int), 
                u=(1,-1,-1), 
                dag={"A":[], "B": [], "C":[], "L":["A", "B", "C"]})
    lamp_exp_causes = [{'B'}, {'C'}]
    scm_lamp.find_causes()
    lamp_causes = scm_lamp.causes
    assert set(map(frozenset, lamp_exp_causes)) == set(map(frozenset, lamp_causes))

def test_example_3():
    suzzy_vars = ("ST", "BT", "SH", "BH", "BS")

    t = .05 
    N = 25
    eps = .25 
    beam_size = 10
    lucb_params = {"a": eps, 
                   "cause_eps": .1, 
                   "non_cause_eps": .1, 
                   "beam_eps": .1, 
                   "max_iter": N, 
                   "verbose": 0, 
                   "init_batch_size": 15,
                   "batch_size": 2,
                   "delta": .05,
                   "beam_size": beam_size,
                   "seed": 0
                   }
    
    noisy_suzzy_avg = SCM(
        V=suzzy_vars,
        U=("st", "bt"),
        D=(0,1),
        u=(1,1),
        dag=None,
        model=AvgRockThrowingModel(suzzy_vars, t, N, seed=0, rng=np.random.default_rng(seed=42))
        )
    
    noisy_suzzy_lucb = SCM(
        V=suzzy_vars,
        U=("st", "bt"),
        D=(0,1),
        u=(1,1),
        dag=None,
        model=LUCBRockThrowingModel(suzzy_vars, t, lucb_params, rng=np.random.default_rng(seed=42))
        )
    noisy_suzzy_avg.find_causes(epsilon=eps, beam_size=beam_size)
    noisy_suzzy_lucb.find_causes(epsilon=eps, beam_size=beam_size)

    exp_causes = [{'ST'}, {'SH'}]
    assert set(map(frozenset, exp_causes)) == set(map(frozenset, noisy_suzzy_avg.causes))
    assert set(map(frozenset, exp_causes)) == set(map(frozenset, noisy_suzzy_lucb.causes))

def test_example_4():
    v = suzzy_example_scm.v

    heuristics = {
        "sum of negative variables": lambda s: sum(s),
        "sum of positive variables": lambda s: len(s) - sum(s),
        "sum of counteractual variables": lambda s: sum([s_val != v_val for s_val, v_val in zip(s, v)]),
        "sum of actual variables": lambda s: sum([s_val == v_val for s_val, v_val in zip(s, v)]),
    }
    exp_causes = [{'ST'}, {'SH'}]
    for desc, psi in heuristics.items():
        system_model = SuzzyCustomPsi(psi)
        scm = SCM(
            V=("ST", "BT", "SH", "BH", "BS"),
            U=("st", "bs"),
            D=(0,1),
            u=(1,1),
            model=system_model,
            dag={"ST":[], "BT":[], "SH":["ST"], "BH":["BT","SH"], "BS":["BH","SH"]}
        )
        scm.find_causes()
        assert set(map(frozenset, exp_causes)) == set(map(frozenset, scm.causes))


import pytest
from actualcauses import suzzy_example_scm

def test_toy_scm_MBS():
    suzzy_example_scm.find_causes(ISI=False)
    mbs_causes = suzzy_example_scm.causes
    expected_causes = [{'SH'}, {'ST'}]
    assert set(map(frozenset, mbs_causes)) == set(map(frozenset, expected_causes))

def test_toy_scm_MBS():
    suzzy_example_scm.find_causes(ISI=True)
    mbs_causes = suzzy_example_scm.causes
    expected_causes = [{'SH'}, {'ST'}]
    assert set(map(frozenset, mbs_causes)) == set(map(frozenset, expected_causes))



from .isi import iterative_identification
from .mbs import beam_search
from .lucb import lucb
from .scm import SCM, suzzy_example_scm
from .system_model import SystemModel, BaseNumpyModel, AverageNumpyModel, LUCBNumpyModel, SuzzyExampleSystemModel

__all__ = ['beam_search', 'iterative_identification', 'SCM', 'lucb', 'SystemModel', 'BaseNumpyModel', 'AverageNumpyModel', 'LUCBNumpyModel', 'SuzzyExampleSystemModel', 'suzzy_example_scm']
__version__ = "1.0.0"


from setuptools import setup, find_packages

setup(
    name="actualcauses",
    version="1.0.0",
    packages=find_packages(),
    install_requires=[
        "tqdm>=4.62.3",
        "numpy>=1.21.0",
    ],
    author="SamuelReyd",
    author_email="samuelreyd@gmail.com",
    description="A package for Actual Causes Identification",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/SamuelReyd/ActualCausesIdentification",
)


from collections.abc import Iterable
import time
from .mbs import beam_search, show_rules
from .isi import iterative_identification
from .system_model import SystemModel, SuzzyExampleSystemModel
from typing import Any

class SCM:
    def __init__(self, V:list[str], U:list[str], D:list[list], u:list,
                 model:SystemModel, dag:list[list]=None, v:list=None):
        """
        SCM object that includes everything needed to execute the algorithms
            V: list of endogenous variables labels (the target must be the last variable)
            U: list of exogenous variables labels
            D: list of discrete domains for the variables in V (in the same order)
            u: context (values of U in the right order)
            model: a SystemModel object that must implement __call__ and evaluate_batch 
            dag (optional): a list of lists where each element i is the set of causal parents of V[i].
            v (optional): a list of values for v. If none is provided, then F(u,[]) is used. 
                Providing v can be useful for stochastic experiments.
        """
        # Mandatory parameters
        self.V = V
        self.U = U
        self.model = model
        if isinstance(D[0], Iterable):
            self.D = D
        else:
            self.D = [D] * len(V)
        self.u = u
        if v is None:
            self.v = self({})
        else:
            self.v = v
            
        # Interventions and evaluation functions
        self.dag = dag
        self.init_vars = dag[V[-1]] if dag is not None else None
        
        self.causes = None
        self.causes_hashable = None
        self.witnesses = None
        self.identification_output = None
        self.interventions = None
        self.identification_time = None
        self.n_calls = None
    
    def __call__(self, e:list[tuple[str,Any]]) -> list[Any]:
        return self.model(self.u, e)
        
    def evaluate_batch(self, E: list[list[tuple[str,Any]]], N:int=1) -> list[tuple[float,float]]:
        return self.model.evaluate_batch(self.u, E, N)

    def get_input(self, base=True):
        if base:
            return self.get_input_beam_search()
        assert self.dag is not None
        return self.get_input_beam_search() | {"dag": self.dag, "PA_T":self.init_vars}

    def get_input_beam_search(self):
        return {"v": self.v[:-1], "V": self.V[:-1], "D": self.D[:-1], "simulation": self.evaluate_batch}

    def find_causes(self, ISI=False, **kwargs):
        t = time.time()
        self.model.n_calls = 0
        if ISI:
            out = iterative_identification(**self.get_input(False), **kwargs)
        else:
            out = beam_search(**self.get_input(), **kwargs)
        self.identification_time = time.time() - t
        self.identification_output = out
        self.causes = [elt[3] for elt in out]
        self.causes_hashable = [tuple(sorted(elt)) for elt in self.causes]
        self.witnesses = [elt[4] for elt in out]
        self.interventions = [elt[0] for elt in out]
        self.n_calls = self.model.n_calls
        
    def show_identification_result(self):
        print(f"Found {len(self.causes)} causes in {self.identification_time:.3f}s with {self.n_calls} model calls\n")
        show_rules(self.identification_output)

suzzy_example_scm = SCM(
    V=("ST", "BT", "SH", "BH", "BS"),
    U=("st", "bs"),
    D=(0,1),
    u=(1,1),
    model=SuzzyExampleSystemModel(),
    dag={"ST":[], "BT":[], "SH":["ST"], "BH":["BT","SH"], "BS":["BH","SH"]}
)